/* Custom styles */
#cy {
  width: 100%;
  height: 600px;
  border: 1px solid #ccc;
  background-color: #fafafa;
  margin-bottom: 20px;
  border-radius: 5px;
  /* Prevent selection of text */
  user-select: none;
}

.zoom-controls {
  margin-bottom: 15px;
}

.zoom-controls button {
  padding: 8px 12px;
}

#editor-container {
  width: 100%;
  height: 600px;
  border: 1px solid #ccc;
  margin-bottom: 20px;
  border-radius: 5px;
}

.btn-group .btn.active {
  background-color: #0d6efd;
  color: white;
}

.highlighted {
  border-color: #dc3545;
  border-width: 4px;
}

#message-area {
  margin-bottom: 15px;
}

#violations-area {
  margin-bottom: 15px;
}

body {
  background-color: #f8f9fa;
}

.form-control {
  min-width: 150px;
}

@media (max-width: 576px) {
  #cy {
    height: 400px;
  }
}
